/****************************************************************************
*
*  rxmmutl.cpp
*
*  Create:  20.11.2001
*  Update:  02.12.2001
*  (C)      Marcel Mller 2001
*
****************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdio.h>
#include <ctype.h>
#include <limits.h>

#ifdef HAVE_CRYPTLIB
#include <md2.h>
#include <md4.h>
#include <md5.h>
#include <crc.h>
#include <ripemd.h>
#include <sha.h>
#include <tiger.h>
#include <haval.h>
#include <adler32.h>
#include <panama.h>
#endif

extern "C"
{
   #define  INCL_WINCOUNTRY
   #define  INCL_BASE
   #define  INCL_REXXSAA
   #include <os2emx.h>
   #define  VALID_ROUTINE   0
   #define  INVALID_ROUTINE 1

   RexxFunctionHandler  MMLoadFuncs;
   RexxFunctionHandler  MMDropFuncs;
   RexxFunctionHandler  MMTranslateCp;
   RexxFunctionHandler  MMUpper;
   RexxFunctionHandler  MMTime;
   RexxFunctionHandler  MMSetFileSize;
   RexxFunctionHandler  MMFileMove;
   RexxFunctionHandler  MMHash;
}


#define BUILDRXSTRING(t, s) memcpy((t)->strptr,(s),(t)->strlength = strlen((s)));

/* Convert REXX string to unsigned long
   return true  conversion succesful
          false error
*/
static bool str2ulong(const RXSTRING str, unsigned long* dst)
{  if (!RXVALIDSTRING(str))
      return false;
   PSZ cp = str.strptr;
   ULONG len = str.strlength;
   *dst = 0;
   for(;;)
   {  if (!isdigit(*cp))
         return false;
      *dst *= 10;
      if (*dst > ULONG_MAX - *cp + '0')
         return false;
      *dst += *cp++ - '0';
      if (!--len)
         return true;
      if (*dst > ULONG_MAX/10)
         return false;
}  }

/* copy REXX string assuming that the destination has storage for 256 bytes of data
   return true  successful
          false out of memory
*/
static bool rxstrcpy(const RXSTRING src, RXSTRING* dst)
{  if ((dst->strlength = src.strlength) <= 256)
      memcpy(dst->strptr, src.strptr, src.strlength);
    else if (DosAllocMem((PPVOID)&dst->strptr, src.strlength, PAG_COMMIT|PAG_WRITE))
   {  BUILDRXSTRING(dst, "Error:");
      return false;
   } else
      memcpy(dst->strptr, src.strptr, src.strlength);
   return true;
}

/* Get HAB for Win... functions */
static HAB hab = (HAB)0;
static bool habd = 0;
static HAB GetHAB()
{  if (hab == (HAB)0 && (habd = !(hab = WinInitialize(0))))// create anchor block
      hab = WinQueryAnchorBlock(HWND_DESKTOP);// get desktop anchor
   return hab;
}
/* free HAB */
static void FreeHAB()
{  if (!habd)
      WinTerminate(hab);
}

/* Translate string from source CP to destination CP
   ARG(1) string
   ARG(2) source CP, current CP by default
   ARG(3) destination CP, current CP by default
   ARG(4) substitution char, 0xFF by default, '' => leave unchanged
*/
ULONG MMTranslateCp(const UCHAR *name, ULONG numargs, RXSTRING args[], const UCHAR *queuename, RXSTRING *retstr)
{  if (numargs < 1 || numargs > 4 || !args[0].strptr)     // check arguments
      return INVALID_ROUTINE;

   ULONG scp = 0, dcp = 0, dummy;
   UCHAR sub = 0xFF;
   bool nosub = false;
   // error char
   if (numargs >= 4 && args[3].strptr)
      if (args[3].strlength > 1)
         return INVALID_ROUTINE;
       else if (args[3].strlength)
         sub = *args[3].strptr;
       else
         nosub = true;
   // source CP
   if (numargs >= 2 && args[1].strptr && !str2ulong(args[1], &scp))
      return INVALID_ROUTINE;
   if (scp == 0)
      DosQueryCp(sizeof(ULONG), &scp, &dummy);// current cp
   // destination CP
   if (numargs >= 3 && args[2].strptr && !str2ulong(args[2], &dcp))
      return INVALID_ROUTINE;
   if (dcp == 0)
      DosQueryCp(sizeof(ULONG), &dcp, &dummy);// current cp
   // duplicate source string
   rxstrcpy(args[0], retstr);
   if (scp == dcp || retstr->strlength == 0)
      return VALID_ROUTINE;
   // translate string
   HAB hab = GetHAB();
   PSZ cp = retstr->strptr + retstr->strlength;
   do
   {  if (*--cp != 0 && *cp != 0xFF)
         if ((*cp = WinCpTranslateChar(hab, scp, *cp, dcp)) == 0)
         {  // Error
            BUILDRXSTRING(retstr, "Error:");
            break;
         } else if (*cp == 0xFF)
         {  if (nosub)
               *cp = args[0].strptr[cp - retstr->strptr];
             else
               *cp = sub;
         }
   } while (cp != retstr->strptr);
   FreeHAB();
   return VALID_ROUTINE;
}

/* Translate string to uppercase using special CP
   ARG(1) string
   ARG(2) CP, current CP by default
   ARG(3) country code, from CONFIG.SYS by default
*/
ULONG MMUpper(const UCHAR *name, ULONG numargs, RXSTRING args[], const UCHAR *queuename, RXSTRING *retstr)
{  if (numargs < 1 || numargs > 3 || !args[0].strptr)     // check arguments
      return INVALID_ROUTINE;

   ULONG scp = 0, cc = 0;
   // CP
   if (numargs >= 2 && args[1].strptr && !str2ulong(args[1], &scp))
      return INVALID_ROUTINE;
   // country code
   if (numargs >= 3 && args[2].strptr && !str2ulong(args[2], &cc))
      return INVALID_ROUTINE;
   // duplicate source string
   rxstrcpy(args[0], retstr);
   if (retstr->strlength == 0)
      return VALID_ROUTINE;
   // translate string
   HAB hab = GetHAB();
   PSZ cp = retstr->strptr + retstr->strlength;
   do
   {  if (*--cp != 0 && (*cp = WinUpperChar(hab, scp, cc, *cp)) == 0)
      {  // Error
         BUILDRXSTRING(retstr, "Error:");
         break;
      }
   } while (cp != retstr->strptr);
   FreeHAB();
   return VALID_ROUTINE;
}

/* Get UNIX-Time
*/
ULONG MMTime(const UCHAR *name, ULONG numargs, RXSTRING args[], const UCHAR *queuename, RXSTRING *retstr)
{  if (numargs > 0)                    // check arguments
      return INVALID_ROUTINE;
   retstr->strlength = sprintf((char*)retstr->strptr, "%lu", time(NULL));
   return VALID_ROUTINE;
}

/* Set file size
   ARG(1) filename
   ARG(2) new size
*/
ULONG MMSetFileSize(const UCHAR *name, ULONG numargs, RXSTRING args[], const UCHAR *queuename, RXSTRING *retstr)
{  ULONG size;
   if (numargs != 2 || !args[0].strptr || !args[1].strptr || !str2ulong(args[1], &size)) // check arguments
      return INVALID_ROUTINE;
   HFILE f;
   ULONG d;
   APIRET r = DosOpen(args[0].strptr, &f, &d, 0, 0, OPEN_ACTION_FAIL_IF_NEW|OPEN_ACTION_OPEN_IF_EXISTS,
    OPEN_SHARE_DENYNONE|OPEN_ACCESS_WRITEONLY, NULL);
   if (r == NO_ERROR)
   {  r = DosSetFileSize(f, size);
      DosClose(f);
   }
   retstr->strlength = sprintf((char*)retstr->strptr, "%lu", r);
   return VALID_ROUTINE;
}

/* Move file or directory
   ARG(1) old filename
   ARG(2) new filename
*/
ULONG MMFileMove(const UCHAR *name, ULONG numargs, RXSTRING args[], const UCHAR *queuename, RXSTRING *retstr)
{  if (numargs != 2 || !args[0].strptr || !args[1].strptr) // check arguments
      return INVALID_ROUTINE;
   APIRET r = DosMove(args[0].strptr, args[1].strptr);
   retstr->strlength = sprintf((char*)retstr->strptr, "%lu", r);
   return VALID_ROUTINE;
}

/* Calculate hash value
   ARG(1) input string
   ARG(2) hash type
   return hashvalue or '' on error
*/
#ifdef HAVE_CRYPTLIB
using namespace CryptoPP;

template <class T>
static HashModule* DefalutHashFactory(ULONG numargs, RXSTRING*)
{  return numargs ? NULL : new T();
}

struct HashEntry
{  const char* name;
   HashModule* (*factory)(ULONG numargs, RXSTRING* args);
} const HashList[] =
#define DEFHE(name) {#name, &DefalutHashFactory<name>}
// the following list MUST be in case insensitive lexical order
{  DEFHE(Adler32),
   DEFHE(CRC32),
   DEFHE(HAVAL),
   DEFHE(MD2),
   DEFHE(MD4),
   DEFHE(MD5),
   DEFHE(RIPEMD160),
   DEFHE(SHA),
   DEFHE(SHA256),
   DEFHE(SHA384),
   DEFHE(SHA512),
   DEFHE(Tiger),
};
#undef DEFHE

static int HashEntryCmp(const char* key, const HashEntry* elem)
{  return stricmp(key, elem->name);
}
#endif

ULONG MMHash(const UCHAR *name, ULONG numargs, RXSTRING args[], const UCHAR *queuename, RXSTRING *retstr)
{  if (numargs < 2 || !args[0].strptr || !args[1].strptr)
      return INVALID_ROUTINE;
   #ifdef HAVE_CRYPTLIB
   const HashEntry* he;
   if ( args[1].strlength > 15
       || (args[1].strptr[args[1].strlength] = 0, // assuming at least 16 bytes of storage for the input buffer
      // hash factory
        he = (HashEntry*)bsearch(args[1].strptr, HashList, sizeof HashList / sizeof *HashList, sizeof *HashList,
	      (int (*)(const void*, const void*))HashEntryCmp)) == NULL )
	   // hash type not found
		return INVALID_ROUTINE;	
   HashModule* hm = he->factory(numargs-2, args+2);
  	if (hm)
	{  // calc hash
	   hm->CalculateDigest(retstr->strptr, args[0].strptr, args[0].strlength); // assuming digestlength < 256
   	retstr->strlength = hm->DigestSize();
	   delete hm;
   	return VALID_ROUTINE;
   }
   #endif
   // some error occured
   retstr->strlength = 0;
   return VALID_ROUTINE;
}

/********** Load all functions */
static PSZ RxFncTable[] =
{  (PSZ)"MMTranslateCp",
   (PSZ)"MMUpper",
   (PSZ)"MMTime",
   (PSZ)"MMSetFileSize",
   (PSZ)"MMFileMove",
   (PSZ)"MMHash",
   (PSZ)"MMDropFuncs",
   (PSZ)"MMLoadFuncs"
};
ULONG MMLoadFuncs(const UCHAR *name, ULONG numargs, RXSTRING args[], const UCHAR *queuename, RXSTRING *retstr)
{  retstr->strlength = 0;              // set return value
   if (numargs > 0)                    // check arguments
      return INVALID_ROUTINE;

   const PSZ* fp = RxFncTable + sizeof(RxFncTable) / sizeof(PSZ) -1;
   do
   {  --fp;
      RexxRegisterFunctionDll(*fp, (PSZ)"RXMMUTL", *fp);
   } while (fp != RxFncTable);
   return VALID_ROUTINE;
}

ULONG MMDropFuncs(const UCHAR *name, ULONG numargs, RXSTRING args[], const UCHAR *queuename, RXSTRING *retstr)
{  retstr->strlength = 0;              // set return value
   if (numargs > 0)                    // check arguments
      return INVALID_ROUTINE;

   const PSZ* fp = RxFncTable + sizeof(RxFncTable) / sizeof(PSZ);
   do RexxDeregisterFunction(*--fp);
    while (fp != RxFncTable);
   return VALID_ROUTINE;
}

